/*
 * Copyright (C) 1989-95 GROUPE BULL
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * GROUPE BULL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of GROUPE BULL shall not be
 * used in advertising or otherwise to promote the sale, use or other dealings
 * in this Software without prior written authorization from GROUPE BULL.
 */

/*****************************************************************************\
*  CrPFrData.c:                                                               *
*                                                                             *
*  XPM library                                                                *
*  Parse an Xpm array and create the pixmap and possibly its mask             *
*                                                                             *
*  Developed by Arnaud Le Hors                                                *
\*****************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include "XpmI.h"

int
XpmCreatePixmapFromData(
    Display		 *display,
    Drawable		  d,
    char		**data,
    Pixmap		 *pixmap_return,
    Pixmap		 *shapemask_return,
    XpmAttributes	 *attributes)
{
    XImage *ximage = NULL, *shapeimage = NULL;
    int ErrorStatus;

    /* initialize return values */
    if (pixmap_return)
	*pixmap_return = 0;
    if (shapemask_return)
	*shapemask_return = 0;

    /* create the images */
    ErrorStatus = XpmCreateImageFromData(display, data,
					 (pixmap_return ? &ximage : NULL),
					 (shapemask_return ?
					  &shapeimage : NULL),
					 attributes);

    if (ErrorStatus != XpmSuccess)
	goto cleanup;

    /* create the pixmaps and destroy images */
    if (pixmap_return && ximage) {
	ErrorStatus =
	    xpmCreatePixmapFromImage(display, d, ximage, pixmap_return);
	if (ErrorStatus < 0)		/* fatal error */
	    goto cleanup;
    }
    if (shapemask_return && shapeimage) {
	ErrorStatus =
	    xpmCreatePixmapFromImage(display, d, shapeimage, shapemask_return);
    }

  cleanup:
    if (ximage != NULL)
	XDestroyImage(ximage);
    if (shapeimage != NULL)
	XDestroyImage(shapeimage);
    if (ErrorStatus < 0) {
	if (pixmap_return && *pixmap_return) {
	    XFreePixmap(display, *pixmap_return);
	    *pixmap_return = 0;
	}
	if (shapemask_return && *shapemask_return) {
	    XFreePixmap(display, *shapemask_return);
	    *shapemask_return = 0;
	}
    }
    return (ErrorStatus);
}
